__title__ = "videohash"
__description__ = "Python package for Perceptual Video Hashing (Near Duplicate Video Detection) - Get a 64-bit comparable hash-value for any video."
__url__ = "https://akamhy.github.io/videohash/"
__version__ = "2.0.2"
__status__ = "production"
__author__ = "Akash Mahanty"
__author_email__ = "akamhy@yahoo.com"
__license__ = "MIT"
__copyright__ = "Copyright 2021 Akash Mahanty"
